#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : EXP
# @Time   : 2020/8/31 10:10
# @File   : _pdm.py
# -----------------------------------------------

import re
import string
from .assist.cfg import *
from .assist.num import *
from .assist import log


class PDM :

    def __init__(self, dbc, pdm_pkg) :
        self.dbc = dbc
        self.pdm_pkg = pdm_pkg
        pdm_path = pdm_pkg.replace('.', '/')
        self.bean_pkg_path = '%s/bean/' % pdm_path
        self.dao_pkg_path = '%s/dao/' % pdm_path
        self.is_mysql = (self.dbc.dbtype() == MYSQL)


    def to_pdm(self, table_whitelist, table_blacklist) :
        conn = self.dbc.conn()
        if not table_whitelist :
            table_whitelist = self._get_all_tables(conn)
        if not table_blacklist :
            table_blacklist = []
        log.info('目标数据表清单：{\n\t%s\n}' % '\n\t'.join(table_whitelist))

        paths = []
        for table_name in table_whitelist :
            if table_name in table_blacklist :
                continue

            columns = self._get_columns(conn, table_name)
            bean_file_content = self._to_beans(table_name, columns)
            path = self._save(bean_file_content, self.bean_pkg_path, table_name, '.py')
            paths.append(path)

            dao_file_content = self._to_daos(table_name, columns)
            path = self._save(dao_file_content, self.dao_pkg_path, table_name, '.py')
            paths.append(path)

        self.dbc.close()
        return paths


    def _get_all_tables(self, conn) :
        tables = []
        try:
            sql = 'show tables' if self.is_mysql else \
                'select name from sqlite_master where type="table" order by name'
            cursor = conn.cursor()
            cursor.execute(sql)
            rows = cursor.fetchall()
            for row in rows:
                if self.is_mysql :
                    table_name = row[0].encode(CHARSET)
                else :
                    table_name = row[0]
                    if table_name == 'sqlite_sequence' :    # sqlite 内置自增序列表
                        continue
                tables.append(byte_to_str(table_name))
            cursor.close()
        except:
            log.error("查询 table 列表失败")
        return tables


    def _get_columns(self, conn, table_name) :
        cursor = conn.cursor()
        cursor.execute('select * from %s limit 0' % table_name)
        columns = [tuple[0] for tuple in cursor.description]
        log.debug('已获取数据表 [%s] 的列名集： %s' % (table_name, columns))
        cursor.close()
        return columns


    def _to_beans(self, table_name, columns) :
        log.debug('正在构造数据表 [%s] 的 Bean 文件' % table_name)
        tpl = DBTemplate(BEAN_TPL)
        variables = list(map(self._to_var, columns))
        placeholders = {
            '{table_name}': table_name,
            '{TableName}': self._to_camel(table_name),
            '{columns}': '\n'.join(list(map((lambda col: '\t%s = "%s"' % (col, col)), columns))),
            '{variables}': '\n'.join(list(map((lambda col: '\t\tself.%s = None' % col), variables))),
            '{params}': '\n'.join(list(('\t\t\tself.%s,' % col) for col in variables[1:])),
            '{kvs}': '\n'.join(list(map(self._to_kv, columns)))
        }
        return tpl.safe_substitute(placeholders).replace('\t', '    ')


    def _to_camel(self, underline) :
        camel = ''
        if isinstance(underline, str) :
            for word in underline.split('_') :
                camel += word.capitalize()
        return camel


    def _to_var(self, col) :
        return re.sub(r'^[a-zA-Z]_', '', col)


    def _to_kv(self, col) :
        return '\t\t\t\t"\t%s = %s" % (self.' + col + ', self.' + self._to_var(col) + '),'


    def _to_daos(self, table_name, columns) :
        log.debug('正在构造数据表 [%s] 的 DAO 文件' % table_name)
        tpl = DBTemplate(DAO_TPL)
        placeholders = {
            # '{pkg_path}': self.pdm_pkg,
            '{table_name}': table_name,
            '{TableName}': self._to_camel(table_name),
            '{insert}': self._to_insert(table_name, columns),
            '{update}': self._to_update(table_name, columns),
            '{select}': self._to_select(table_name, columns),
            '{set_values}': '\n'.join(('\t\t\tbean.%s = self._to_val(row, %i)' % (self._to_var(col), idx)) for idx, col in enumerate(columns))
        }
        return tpl.safe_substitute(placeholders).replace('\t', '    ')


    def _to_insert(self, table_name, columns) :
        cols = ', '.join(columns[1:])
        if self.is_mysql :
            vars = 's, %' * (len(columns) - 2)
            sql = 'insert into ' + table_name + '(' + cols + ') values(%' + vars + 's)'
        else :
            vars = ', ?' * (len(columns) - 2)
            sql = 'insert into ' + table_name + '(' + cols + ') values(?' + vars + ')'
        return sql


    def _to_update(self, table_name, columns) :
        cols = ', '.join(list(map(
            (lambda col: col + ' = ' + ('%s' if self.is_mysql else '?')),
            columns[1:]
        )))
        return 'update ' + table_name + ' set ' + cols + ' where 1 = 1 '


    def _to_select(self, table_name, columns) :
        sql = 'select %s from ' + table_name + ' where 1 = 1 '
        return sql % (', '.join(columns))


    def _save(self, content, filedir, filename, suffix) :
        '''
        写入 PDM 文件
        :param content: 文件内容
        :param filedir: 存储目录
        :param filename: 文件名
        :param suffix: 文件后缀
        :return: PDM 文件路径
        '''
        if not os.path.exists(filedir) :
            os.makedirs(filedir)
            self._create_init_py(filedir)

        path = '%s%s%s' % (filedir, filename, suffix)
        with open(path, 'w+', encoding=CHARSET) as file :
            file.write(content)
        return path


    def _create_init_py(self, pkg_dir) :
        '''
        在包目录的每一层创建 __init__.py 文件
        :param pkg_dir: 包目录
        '''
        dir = pkg_dir
        while dir :
            initpy_path = '%s/__init__.py' % dir
            with open(initpy_path, 'w+', encoding=CHARSET) as file : file.write('')
            dir = os.path.dirname(dir)



class DBTemplate(string.Template) :
    delimiter = '@'
    idpattern = r'\{\w+\}'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -------------------------------
# 数据访问对象：基类
# -------------------------------


from ..assist import log
from ..assist.cfg import *


class BaseDao:
    """
    Dao 基类
    """

    # TODO 需子类实现
    TABLE_NAME = ""
    SQL_COUNT = ""
    SQL_TRUNCATE = ""
    SQL_INSERT = ""
    SQL_DELETE = ""
    SQL_UPDATE = ""
    SQL_SELECT = ""
    CHARSET = "utf8"

    def __init__(self):
        pass


    def count(self, conn):
        """
        统计行数
        :param conn: 数据库连接
        :return: 表行数
        """
        cnt = 0
        try:
            cursor = conn.cursor()
            cursor.execute(self.SQL_COUNT)
            cnt = cursor.fetchone()[0]
            cursor.close()
        except:
            log.error("统计表 [%s] 行数失败" % self.TABLE_NAME)
        return cnt


    def truncate(self, conn):
        """
        清空表
        :param conn: 数据库连接
        :return: 是否清空成功
        """
        is_ok = False
        try:
            cursor = conn.cursor()
            cursor.execute(self.SQL_TRUNCATE if conn.dbtype() == MYSQL else self.SQL_DELETE)
            conn.commit()
            cursor.close()
            is_ok = True
        except:
            log.error("清空表 [%s] 失败" % self.TABLE_NAME)
        return is_ok


    def insert(self, conn, bean):
        """
        插入单条数据
        :param conn: 数据库连接
        :param bean: 数据模型实例
        :return: 是否插入成功
        """
        is_ok = False
        try:
            cursor = conn.cursor()
            params = bean.params()
            cursor.execute(self.SQL_INSERT, params)
            conn.commit()
            cursor.close()
            is_ok = True
        except:
            log.error("插入数据到表 [%s] 失败" % self.TABLE_NAME)
        return is_ok


    def insert_all(self, conn, beans):
        """
        插入多条数据
        :param conn: 数据库连接
        :param beans: 数据模型实例队列
        :return: 成功插入个数
        """
        cnt = 0
        try:
            cursor = conn.cursor()
            for bean in beans:
                try:
                    params = bean.params()
                    cursor.execute(self.SQL_INSERT, params)
                    cnt += 1
                except:
                    log.error("插入数据到表 [%s] 失败" % self.TABLE_NAME)
            conn.commit()
            cursor.close()
        except:
            log.error("插入数据集到表 [%s] 失败" % self.TABLE_NAME)
        return cnt


    def delete(self, conn, wheres={}):
        """
        删除数据
        :param conn: 数据库连接
        :param wheres: 条件键值对, 要求键值包含操作符，如： { 'column1 like': 'xyz', 'column2 =': 'abc' }
        :return: 是否删除成功
        """
        is_ok = False
        try:
            cursor = conn.cursor()
            sql = self._append(conn.dbtype(), self.SQL_DELETE, wheres.keys())
            cursor.execute(sql, list(wheres.values()))
            conn.commit()
            cursor.close()
            is_ok = True
        except:
            log.error("从表 [%s] 删除数据失败" % self.TABLE_NAME)
        return is_ok


    def update(self, conn, bean):
        """
        更新数据
        :param conn: 数据库连接
        :param bean: 数据模型实例
        :return: 是否更新成功
        """
        is_ok = False
        try:
            cursor = conn.cursor()
            sql = self._append(conn.dbtype(), self.SQL_UPDATE, [bean.i_id + ' = '])
            params = bean.params() + (bean.id,)
            cursor.execute(sql, params)
            conn.commit()
            cursor.close()
            is_ok = True
        except:
            log.error("更新数据到表 [%s] 失败" % self.TABLE_NAME)
        return is_ok


    def query_all(self, conn):
        """
        查询表中所有数据
        :param conn: 数据库连接
        :return: 数据模型实例队列（失败返回 [] ，不会为 None）
        """
        return self.query_some(conn)


    def query_some(self, conn, wheres={}):
        """
        查询表中部分数据
        :param conn: 数据库连接
        :param wheres: 条件键值对, 要求键值包含操作符，如： { 'column1 like': 'xyz', 'column2 =': 'abc' }
        :return: 数据模型实例队列（失败返回 [] ，不会为 None）
        """
        beans = []
        try:
            cursor = conn.cursor()
            sql = self._append(conn.dbtype(), self.SQL_SELECT, wheres.keys())
            cursor.execute(sql, list(wheres.values()))
            rows = cursor.fetchall()
            for row in rows:
                bean = self._to_bean(row)
                beans.append(bean)
            cursor.close()
        except:
            log.error("从表 [%s] 查询数据失败" % self.TABLE_NAME)
        return beans


    def query_one(self, conn, wheres={}):
        """
        查询表中一条数据
        :param conn: 数据库连接
        :param wheres: 条件键值对, 要求键值包含操作符，如： { 'column1 like': 'xyz', 'column2 =': 'abc' }
        :return: 数据模型实例队列（若多个满足则返回第 1 个，没有满足则返回 None）
        """
        bean = None
        try:
            cursor = conn.cursor()
            sql = self._append(conn.dbtype(), self.SQL_SELECT, wheres.keys())
            cursor.execute(sql, list(wheres.values()))
            row = cursor.fetchone()
            bean = self._to_bean(row)
            conn.commit()
            cursor.close()
        except:
            log.error("从表 [%s] 查询数据失败" % self.TABLE_NAME)
        return bean


    def _append(self, dbtype, sql, keys):
        """
        追加 where 条件到 sql, 条件之间只为 and 关系（目的只是支持简单的数据库操作）
        :param dbtype: 数据库类型
        :param sql: SQL 语句
        :param keys: 条件键值集合, 要求键值包含操作符，如： [ 'column1 like'， 'column2 =' ]
        :return: 追加 where 条件后的 sql
        """
        placeholder = '%s' if dbtype == MYSQL else '?'
        _sql = sql
        if keys:
            for key in keys:
                _sql = " ".join((_sql, "and", key, placeholder))  # %s 是 sql 占位符，目的是防注入
        return _sql


    def _to_bean(self, row):
        """
        把数据库查询的单行结果转换成模型实例对象
        :param row: 单行查询结果
        :return: 模型实例对象
        """
        # 需子类实现
        return row


    def _to_val(self, row, idx):
        """
        把 unicode 编码的字符串转换成 utf8
        :param row: 行对象
        :param idx: 列索引
        :return: 列值（utf8 编码）
        """
        val = None
        try:
            val = row[idx]
            if val is not None and isinstance(val, str):
                val = val.encode(CHARSET)
        except:
            pass
        return val



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : EXP
# -----------------------------------------------


def byte_to_str(value) :
    """
    byte 转 str。
    :param value: byte 值
    :return: str 字符串
    """
    if isinstance(value, bytes) :
        value = bytes.decode(value)
    return value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : EXP
# @Time   : 2020/4/26 13:14
# @File   : log.py
# -----------------------------------------------

import os
import traceback
import logging
from logging.handlers import TimedRotatingFileHandler
from .cfg import PRJ_DIR


IS_INIT = False
LOG_DIR = '%s/log' % PRJ_DIR
RUN_LOG = '%s/run.log' % LOG_DIR
ERR_LOG = '%s/err.log' % LOG_DIR


def init(runlog = RUN_LOG, errlog = ERR_LOG):
    """
    初始化日志配置 （只需在程序入口调用一次）
    :return: None
    """
    if not os.path.exists(LOG_DIR) :
        os.mkdir(LOG_DIR)

    # 全局配置
    logger = logging.getLogger()
    logger.setLevel("DEBUG")
    BASIC_FORMAT = "%(asctime)s [%(levelname)s] : %(message)s"
    DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
    formatter = logging.Formatter(BASIC_FORMAT, DATE_FORMAT)

    # 输出到控制台的 handler
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    ch.setLevel("DEBUG")
    logger.addHandler(ch)

    # 输出到运行日志文件的 handler
    fh = TimedRotatingFileHandler(filename=runlog, when="D", interval=1, backupCount=7)
    fh.setFormatter(formatter)
    fh.setLevel("INFO")
    logger.addHandler(fh)

    # 输出到异常日志文件的 handler
    exfh = TimedRotatingFileHandler(filename=errlog, when="D", interval=1, backupCount=7)
    exfh.setLevel("ERROR")
    exfh.setFormatter(formatter)
    logger.addHandler(exfh)

    global IS_INIT
    IS_INIT = True



def debug(msg):
    """
    打印调试信息
    :param msg: 日志信息
    :return: None
    """
    logging.debug(msg) if IS_INIT else print(msg)


def info(msg):
    """
    打印正常信息
    :param msg: 日志信息
    :return: None
    """
    logging.info(msg) if IS_INIT else print(msg)


def warn(msg):
    """
    打印警告信息
    :param msg: 日志信息
    :return: None
    """
    logging.warning(msg) if IS_INIT else print(msg)


def error(msg):
    """
    打印异常信息和异常堆栈
    :param msg: 日志信息
    :return: None
    """
    if IS_INIT :
        logging.exception(msg)
        logging.exception(traceback.format_exc())
    else :
        print(msg)
        print(traceback.format_exc())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : EXP
# @Time   : 2020/4/28 21:56
# @File   : env.py
# -----------------------------------------------

import os
PRJ_DIR = os.path.dirname(
    os.path.dirname(
        os.path.dirname(
            os.path.dirname(
                os.path.abspath(__file__)
            )
        )
    )
)

CHARSET = 'utf-8'
ENCODING = 'utf8'
MYSQL = 'mysql'
SQLITE = 'sqlite'


BEAN_TPL = '''#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -------------------------------
# PDM: @{table_name}
# -------------------------------

class @{TableName} :
    table_name = '@{table_name}'
@{columns}


    def __init__(self) :
@{variables}


    def params(self) :
        return (
@{params}
        )


    def __repr__(self) :
        return '\\n'.join(
            (
                '%s: {' % self.table_name,
@{kvs}
                '}\\n'
            )
        )
'''


DAO_TPL = '''#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -------------------------------
# DAO: @{table_name}
# -------------------------------

from ..bean.@{table_name} import @{TableName}
from pypdm.dao._base import BaseDao


class @{TableName}Dao(BaseDao) :
    TABLE_NAME = '@{table_name}'
    SQL_COUNT = 'select count(1) from @{table_name}'
    SQL_TRUNCATE = 'truncate table @{table_name}'
    SQL_INSERT = '@{insert}'
    SQL_DELETE = 'delete from @{table_name} where 1 = 1 '
    SQL_UPDATE = '@{update}'
    SQL_SELECT = '@{select}'

    def __init__(self) :
        BaseDao.__init__(self)

    def _to_bean(self, row) :
        bean = None
        if row:
            bean = @{TableName}()
@{set_values}
        return bean
'''


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : EXP
# @Time   : 2020/4/29 22:14
# @File   : _sqlite.py
# -------------------------------
# Mysql 数据库连接器
# -------------------------------

import pymysql as pdbc
from ..assist.cfg import *
from ..assist import log


class MysqlDBC :
    """
    Mysql 数据库连接器
    """

    def __init__(self, host='127.0.0.1', port=3306, username='root', password='123456', dbname='test', encoding=ENCODING, options={}) :
        """
        构造函数
        :param host : 数据库地址
        :param port : 数据库端口
        :param username : 数据库账号
        :param password : 数据库密码
        :param dbname : 数据库名称
        :param encoding : 数据库编码
        :param options : 上述所有数据库参数的字典，方便传参
        """
        self.host = host or options.get('host') or '127.0.0.1'
        self.port = port or options.get('port') or 3306
        self.username = username or options.get('username') or 'root'
        self.password = password or options.get('password') or '123456'
        self.dbname = dbname or options.get('dbname') or 'test'
        self.encoding = encoding or options.get('encoding') or ENCODING
        self.encoding = (ENCODING if encoding.lower() == CHARSET else encoding)
        self._conn = None


    def dbtype(self) :
        '''
        获取数据库类型
        :return: 数据库类型
        '''
        return MYSQL


    def conn(self) :
        """
        连接到数据库
        :return : 数据库连接（失败返回 None）
        """
        if not self._conn :
            try :
                self._conn = pdbc.connect(
                    host = self.host,
                    port = self.port,
                    user = self.username,
                    password = self.password,
                    db = self.dbname,
                    charset = self.encoding
                )
            except :
                log.error("连接数据库 [%s] 失败" % self.dbname)
        return self._conn


    def close(self) :
        """
        断开数据库连接
        :return : 是否断开成功
        """
        is_ok = False
        if self._conn :
            try :
                self._conn.close()
                self._conn = None
                is_ok = True
            except :
                log.error("断开数据库 [%s] 失败" % self.dbname)
        return is_ok


    def reconn(self) :
        """
        重连数据库
        :return : 数据库连接（失败返回 None）
        """
        self.close()
        return self.conn()


    def commit(self) :
        """
        提交事务
        :return : 是否提交成功
        """
        is_ok = False
        if self._conn :
            try :
                self._conn.commit()
                is_ok = True
            except :
                log.error("提交事务到数据库 [%s] 失败" % self.dbname)
        return is_ok


    def cursor(self) :
        """
        返回数据库连接游标
        :return: 游标（失败返回 None）
        """
        return self._conn.cursor() if self._conn else None


    def exec_script(self, filepath) :
        """
        执行 SQL 脚本
        :param filepath : 脚本文件
        :return : 是否执行成功
        """
        is_ok = False
        if self.conn() :
            try :
                with open(filepath, "r", encoding=CHARSET) as file :
                    data = file.read()

                    cursor = self._conn.cursor()
                    sqls = data.split(";")
                    for sql in sqls :
                        sql = sql.strip()
                        if sql :
                            cursor.execute(sql)
                    self._conn.commit()
                    cursor.close()
                    is_ok = True
            except :
                log.error("执行 SQL 脚本失败： [%s]" % filepath)
            self.close()
        return is_ok


    def exec_sql(self, sql):
        """
        执行 SQL 语句
        :param sql: SQL 语句
        :return: 是否执行成功
        """
        is_ok = False
        if self.conn() :
            try:
                cursor = self._conn.cursor()
                cursor.execute(sql)
                self._conn.commit()
                cursor.close()
                is_ok = True
            except:
                log.error("执行 SQL 失败： [%s]" % sql)
        return is_ok

